from vulkan import *
import glfw
from glfw.GLFW import GLFW_NO_API, GLFW_CLIENT_API, GLFW_RESIZABLE, GLFW_FALSE

reqExtensions = [VK_KHR_SWAPCHAIN_EXTENSION_NAME, ]


# trieda QueueFamilyIndices ------------------------------------------------------------------------
class QueueFamilyIndices:
    def __init__(self):
        self.graphicsFamily = None  # graphics queue family index
        self.presentFamily = None   # present queue family index

    def getState(self):
        return (self.graphicsFamily is not None) and (self.presentFamily is not None)

    def getQueues(self):
        return [self.graphicsFamily, self.presentFamily]


# trieda SwapchainSupportDetails -------------------------------------------------------------------
class SwapchainSupportDetails:
    def __init__(self):
        self.capabilities = None
        self.formats = []
        self.presentModes = []


# trieda VulkanApp ---------------------------------------------------------------------------------
class VulkanApp:
    # konstruktor
    def __init__(self):
        self.width = 800                        # sirka okna
        self.height = 600                       # vyska okna
        self.window = None                      # GLFW okno
        self.instance = None                    # izolovana instancia API Vulkan
        self.surface = None                     # WIN surface VkSurfaceKHR
        self.physicalDevice = VK_NULL_HANDLE    # Fyzicke zariadenie
        self.logicalDevice = None               # Logicke zariadenie
        self.graphicsQueue = None               # Graphics Queue
        self.presentQueue = None                # Present Queue
        self.swapchain = None                   # Swapchain
        self.swapchainImages = None             # Swapchain Images
        self.swapchainImageFormat = None        # Swapchain ImageFormat
        self.swapchainExtent = None             # Swapchain Extent
        self.swapchainImageViews = []           # Swapchain ImageViews
        self.renderPass = None                  # RenderPass
        self.swapchainFramebuffers = []         # Swapchain Framebuffers
        self.pipelineLayout = None              # GraphicsPipelineLayout
        self.graphicsPipeline = None            # GraphicsPipeline
        self.commandPool = None                 # Command Pool
        self.commandBuffers = None              # Command Buffers
        self.imageAvailableSemaphore = None     # Semaphore
        self.renderFinishedSemaphore = None     # Semaphore

    # Vytvorenie GLFW okna -------------------------------------------------------------------------
    def createWindow(self):
        glfw.init()
        glfw.window_hint(GLFW_CLIENT_API, GLFW_NO_API)
        glfw.window_hint(GLFW_RESIZABLE, GLFW_FALSE)

        self.window = glfw.create_window(self.width, self.height, "PyGLSL v2.0", None, None)

    # Ziskanie indexov graphicsFamily a presentFamily Queues ---------------------------------------
    def getQueueFamilyIndices(self, device):
        indices = QueueFamilyIndices()  # instancia triedy QueueFamilyIndices
        vkGetPhysicalDeviceSurfaceSupportKHR = vkGetInstanceProcAddr(self.instance,
                                                                     'vkGetPhysicalDeviceSurfaceSupportKHR')
        queueFamilies = vkGetPhysicalDeviceQueueFamilyProperties(device)
        i = 0
        for q in queueFamilies:
            if q.queueFlags & VK_QUEUE_GRAPHICS_BIT:
                indices.graphicsFamily = i
            if vkGetPhysicalDeviceSurfaceSupportKHR(device, i, self.surface):
                indices.presentFamily = i
            i += 1
        return indices

    # Ziskanie detailov podpory Swapchain ----------------------------------------------------------
    def getSwapchainSupportDetails(self, device):
        details = SwapchainSupportDetails()  # instancia triedy SwapchainSupportDetails
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR = vkGetInstanceProcAddr(self.instance,
                                                                          "vkGetPhysicalDeviceSurfaceCapabilitiesKHR")
        vkGetPhysicalDeviceSurfaceFormatsKHR = vkGetInstanceProcAddr(self.instance,
                                                                     "vkGetPhysicalDeviceSurfaceFormatsKHR")
        vkGetPhysicalDeviceSurfacePresentModesKHR = vkGetInstanceProcAddr(self.instance,
                                                                          "vkGetPhysicalDeviceSurfacePresentModesKHR")
        details.capabilities = vkGetPhysicalDeviceSurfaceCapabilitiesKHR(device, self.surface)
        details.formats = vkGetPhysicalDeviceSurfaceFormatsKHR(device, self.surface)
        details.presentModes = vkGetPhysicalDeviceSurfacePresentModesKHR(device, self.surface)
        return details

    # Kontrola pouzitelnosti fyzickeho zariadenia --------------------------------------------------
    def isPhysicalDeviceSuitable(self, device):
        # Ziskanie indexov graphicsFamily a presentFamily Queues
        indices = self.getQueueFamilyIndices(device)
        supExtensions = vkEnumerateDeviceExtensionProperties(device, None)
        for e in reqExtensions:
            extSupported = False
            for f in supExtensions:
                if e == f.extensionName:
                    extSupported = True
            if not extSupported:
                return False
        # Ziskanie detailov podpory Swapchain
        swapchainDetails = self.getSwapchainSupportDetails(device)
        return indices.getState() and \
               len(swapchainDetails.formats) != 0 and \
               len(swapchainDetails.presentModes) != 0

    # Vytvorenie Shader modulu ---------------------------------------------------------------------
    def createShaderModule(self, fpath):
        with open(fpath, 'rb') as file:
            code = file.read()
            ShaderModuleCreateInfo = VkShaderModuleCreateInfo(
                sType=VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
                codeSize=len(code),
                pCode=code
            )
            return vkCreateShaderModule(self.logicalDevice, ShaderModuleCreateInfo, None)

    # Inicializacia API Vulkan ---------------------------------------------------------------------
    def initVulkan(self):
        # Vytvorenie izolovanej instancie API Vulkan -----------------------------------------------
        # ApplicationInfo
        ApplicationInfo = VkApplicationInfo(
            sType=VK_STRUCTURE_TYPE_APPLICATION_INFO,
            pApplicationName="PyGLSL v2.0",
            applicationVersion=VK_MAKE_VERSION(2, 0, 0),
            pEngineName="N/A",
            engineVersion=VK_MAKE_VERSION(1, 0, 0),
            apiVersion=VK_API_VERSION_1_0)
        # Vulkan instance extensions potrebne pre GLFW
        Extensions = glfw.get_required_instance_extensions()
        # InstanceCreateInfo
        InstanceCreateInfo = VkInstanceCreateInfo(
            sType=VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
            pApplicationInfo=ApplicationInfo,
            enabledExtensionCount=len(Extensions),
            ppEnabledExtensionNames=Extensions,
            enabledLayerCount=0,
            ppEnabledLayerNames=None)
        try:
            self.instance = vkCreateInstance(InstanceCreateInfo, None)
        except Exception:
            raise RuntimeError("Chyba vytvorenia instancie API Vulkan!")

        # Vytvorenie WIN surface -------------------------------------------------------------------
        surface_ptr = ffi.new('VkSurfaceKHR[1]')
        glfw.create_window_surface(self.instance, self.window, None, surface_ptr)
        self.surface = surface_ptr[0]
        if self.surface is None:
            raise RuntimeError("Chyba vytvorenia WIN Surface!")

        # Vyber fyzickeho zariadenia - GPU ---------------------------------------------------------
        devices = vkEnumeratePhysicalDevices(self.instance)
        for d in devices:
            if self.isPhysicalDeviceSuitable(d):  # Kontrola pouzitelnosti fyzickeho zariadenia
                self.physicalDevice = d
                break
        if self.physicalDevice == VK_NULL_HANDLE:
            raise RuntimeError("Nemozno vybrat pouzitelne fyzicke zariadenie - GPU!")
        GPUindices = self.getQueueFamilyIndices(self.physicalDevice)

        # Vytvorenie logickeho zariadenia ----------------------------------------------------------
        DeviceQueueCreateInfo = []
        for q in GPUindices.getQueues():
            DeviceQueueCreateInfo.append(VkDeviceQueueCreateInfo(
                sType=VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
                queueFamilyIndex=q,
                queueCount=1,
                pQueuePriorities=[1, ])
            )
        PhysicalDeviceFeatures = VkPhysicalDeviceFeatures()
        DeviceCreateInfo = VkDeviceCreateInfo(
            sType=VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
            flags=0,
            pQueueCreateInfos=DeviceQueueCreateInfo,
            queueCreateInfoCount=len(DeviceQueueCreateInfo),
            pEnabledFeatures=PhysicalDeviceFeatures,
            enabledExtensionCount=len(reqExtensions),
            ppEnabledExtensionNames=reqExtensions,
            enabledLayerCount=0,
            ppEnabledLayerNames=None
        )
        try:
            self.logicalDevice = vkCreateDevice(self.physicalDevice, DeviceCreateInfo, None)
        except Exception:
            raise RuntimeError("Chyba vytvorenia logickeho zariadenia!")
        # Graphics a Present Queues
        self.graphicsQueue = vkGetDeviceQueue(self.logicalDevice, GPUindices.graphicsFamily, 0)
        self.presentQueue = vkGetDeviceQueue(self.logicalDevice, GPUindices.presentFamily, 0)

        # Vytvorenie Swapchain ---------------------------------------------------------------------
        swapchainDetails = self.getSwapchainSupportDetails(self.physicalDevice)
        imgCount = swapchainDetails.capabilities.minImageCount + 1
        if (swapchainDetails.capabilities.maxImageCount > 0) and \
                (imgCount > swapchainDetails.capabilities.maxImageCount):
            imgCount = swapchainDetails.capabilities.maxImageCount
        surfaceFormat = swapchainDetails.formats[0]
        for f in swapchainDetails.formats:
            if f.format == VK_FORMAT_B8G8R8A8_SRGB and f.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR:
                surfaceFormat = f
                break
        presentMode = VK_PRESENT_MODE_FIFO_KHR  # double-buffering
        for p in swapchainDetails.presentModes:
            if p == VK_PRESENT_MODE_MAILBOX_KHR:
                presentMode = p                 # triple-buffering
                break
        w = max(swapchainDetails.capabilities.minImageExtent.width,
                min(swapchainDetails.capabilities.maxImageExtent.width, self.width))
        h = max(swapchainDetails.capabilities.minImageExtent.height,
                min(swapchainDetails.capabilities.maxImageExtent.height, self.height))
        surfaceExtent = VkExtent2D(w, h)
        if GPUindices.graphicsFamily != GPUindices.presentFamily:
            cI_imageSharingMode = VK_SHARING_MODE_CONCURRENT
            cI_queueFamilyIndexCount = 2
            cI_pQueueFamilyIndices = [GPUindices.graphicsFamily, GPUindices.presentFamily]
        else:
            cI_imageSharingMode = VK_SHARING_MODE_EXCLUSIVE
            cI_queueFamilyIndexCount = 0
            cI_pQueueFamilyIndices = None
        SwapchainCreateInfo = VkSwapchainCreateInfoKHR(
            sType=VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
            surface=self.surface,
            minImageCount=imgCount,
            imageFormat=surfaceFormat.format,
            imageColorSpace=surfaceFormat.colorSpace,
            imageExtent=surfaceExtent,
            imageArrayLayers=1,
            imageUsage=VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
            imageSharingMode=cI_imageSharingMode,
            queueFamilyIndexCount=cI_queueFamilyIndexCount,
            pQueueFamilyIndices=cI_pQueueFamilyIndices,
            preTransform=swapchainDetails.capabilities.currentTransform,
            compositeAlpha=VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,
            presentMode=presentMode,
            clipped=VK_TRUE,
            oldSwapchain=VK_NULL_HANDLE
        )
        vkCreateSwapchainKHR = vkGetDeviceProcAddr(self.logicalDevice, 'vkCreateSwapchainKHR')
        self.swapchain = vkCreateSwapchainKHR(self.logicalDevice, SwapchainCreateInfo, None)
        vkGetSwapchainImagesKHR = vkGetDeviceProcAddr(self.logicalDevice, 'vkGetSwapchainImagesKHR')
        self.swapchainImages = vkGetSwapchainImagesKHR(self.logicalDevice, self.swapchain)
        self.swapchainImageFormat = surfaceFormat.format
        self.swapchainExtent = surfaceExtent

        # Vytvorenie ImageViews --------------------------------------------------------------------
        for i in self.swapchainImages:
            ImageViewCreateInfo = VkImageViewCreateInfo(
                sType=VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
                image=i,
                viewType=VK_IMAGE_VIEW_TYPE_2D,
                format=self.swapchainImageFormat,
                components=VkComponentMapping(r=VK_COMPONENT_SWIZZLE_IDENTITY, g=VK_COMPONENT_SWIZZLE_IDENTITY,
                                              b=VK_COMPONENT_SWIZZLE_IDENTITY, a=VK_COMPONENT_SWIZZLE_IDENTITY),
                subresourceRange=VkImageSubresourceRange(aspectMask=VK_IMAGE_ASPECT_COLOR_BIT,
                                                         baseMipLevel=0, levelCount=1,
                                                         baseArrayLayer=0, layerCount=1)
            )
            self.swapchainImageViews.append(vkCreateImageView(self.logicalDevice, ImageViewCreateInfo, None))

        # Vytvorenie RederPass ---------------------------------------------------------------------
        colorAttachment = VkAttachmentDescription(
            format=self.swapchainImageFormat,
            samples=VK_SAMPLE_COUNT_1_BIT,
            loadOp=VK_ATTACHMENT_LOAD_OP_CLEAR,
            storeOp=VK_ATTACHMENT_STORE_OP_STORE,
            stencilLoadOp=VK_ATTACHMENT_LOAD_OP_DONT_CARE,
            stencilStoreOp=VK_ATTACHMENT_STORE_OP_DONT_CARE,
            initialLayout=VK_IMAGE_LAYOUT_UNDEFINED,
            finalLayout=VK_IMAGE_LAYOUT_PRESENT_SRC_KHR
        )
        colorAttachmentRef = VkAttachmentReference(
            attachment=0,
            layout=VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL
        )
        subpassDescription = VkSubpassDescription(
            pipelineBindPoint=VK_PIPELINE_BIND_POINT_GRAPHICS,
            colorAttachmentCount=1,
            pColorAttachments=colorAttachmentRef
        )
        subpassDependency = VkSubpassDependency(
            srcSubpass=VK_SUBPASS_EXTERNAL,
            dstSubpass=0,
            srcStageMask=VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            srcAccessMask=0,
            dstStageMask=VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            dstAccessMask=VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT
        )
        RenderPassCreateInfo = VkRenderPassCreateInfo(
            sType=VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
            attachmentCount=1,
            pAttachments=colorAttachment,
            subpassCount=1,
            pSubpasses=subpassDescription,
            dependencyCount=1,
            pDependencies=subpassDependency
        )
        self.renderPass = vkCreateRenderPass(self.logicalDevice, RenderPassCreateInfo, None)

        # Vytvorenie Framebuffers ------------------------------------------------------------------
        for v in self.swapchainImageViews:
            attachments = [v, ]
            FramebufferCreateInfo = VkFramebufferCreateInfo(
                sType=VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
                renderPass=self.renderPass,
                attachmentCount=1,
                pAttachments=attachments,
                width=self.swapchainExtent.width,
                height=self.swapchainExtent.height,
                layers=1
            )
            self.swapchainFramebuffers.append(vkCreateFramebuffer(self.logicalDevice, FramebufferCreateInfo, None))

        # Vytvorenie Shader modulov ----------------------------------------------------------------
        VSModule = self.createShaderModule("shaders/vert.spv")
        VSPipelineShaderStageCreateInfo = VkPipelineShaderStageCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
            stage=VK_SHADER_STAGE_VERTEX_BIT,
            module=VSModule,
            pName="main"
        )
        FSModule = self.createShaderModule("shaders/frag.spv")
        FSPipelineShaderStageCreateInfo = VkPipelineShaderStageCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
            stage=VK_SHADER_STAGE_FRAGMENT_BIT,
            module=FSModule,
            pName="main"
        )
        shaderStages = [VSPipelineShaderStageCreateInfo, FSPipelineShaderStageCreateInfo]

        # Vytvorenie GraphicsPipeline --------------------------------------------------------------
        # VS binding
        PipelineVertexInputStateCreateInfo = VkPipelineVertexInputStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO,
            vertexBindingDescriptionCount=0,
            vertexAttributeDescriptionCount=0
        )
        # assembly
        PipelineInputAssemblyStateCreateInfo = VkPipelineInputAssemblyStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO,
            topology=VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST,
            primitiveRestartEnable=VK_FALSE
        )
        # viewport & scissor
        viewport = VkViewport(
            x=0,
            y=0,
            width=self.swapchainExtent.width,
            height=self.swapchainExtent.height,
            minDepth=0.0,
            maxDepth=1.0
        )
        scissor = VkRect2D(
            offset=[0, 0],
            extent=self.swapchainExtent
        )
        PipelineViewportStateCreateInfo = VkPipelineViewportStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO,
            viewportCount=1,
            pViewports=viewport,
            scissorCount=1,
            pScissors=scissor
        )
        # rasterizer
        PipelineRasterizationStateCreateInfo = VkPipelineRasterizationStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO,
            depthClampEnable=VK_FALSE,
            rasterizerDiscardEnable=VK_FALSE,
            polygonMode=VK_POLYGON_MODE_FILL,
            lineWidth=1.0,
            cullMode=VK_CULL_MODE_BACK_BIT,
            frontFace=VK_FRONT_FACE_CLOCKWISE,
            depthBiasEnable=VK_FALSE
        )
        # multisampling
        PipelineMultisampleStateCreateInfo = VkPipelineMultisampleStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO,
            sampleShadingEnable=VK_FALSE,
            rasterizationSamples=VK_SAMPLE_COUNT_1_BIT
        )
        # color blending
        colorBlendAttachment = VkPipelineColorBlendAttachmentState(
            colorWriteMask=VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT |
            VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT,
            blendEnable=VK_FALSE
        )
        PipelineColorBlendStateCreateInfo = VkPipelineColorBlendStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO,
            logicOpEnable=VK_FALSE,
            attachmentCount=1,
            pAttachments=colorBlendAttachment,
            blendConstants=[0.0, 0.0, 0.0, 0.0]
        )
        PipelineLayoutCreateInfo = VkPipelineLayoutCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO
        )
        self.pipelineLayout = vkCreatePipelineLayout(self.logicalDevice, PipelineLayoutCreateInfo, None)
        GraphicsPipelineCreateInfo = VkGraphicsPipelineCreateInfo(
            sType=VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO,
            stageCount=2,
            pStages=shaderStages,
            pVertexInputState=PipelineVertexInputStateCreateInfo,
            pInputAssemblyState=PipelineInputAssemblyStateCreateInfo,
            pViewportState=PipelineViewportStateCreateInfo,
            pRasterizationState=PipelineRasterizationStateCreateInfo,
            pMultisampleState=PipelineMultisampleStateCreateInfo,
            pDepthStencilState=None,
            pColorBlendState=PipelineColorBlendStateCreateInfo,
            layout=self.pipelineLayout,
            renderPass=self.renderPass,
            subpass=0
        )
        self.graphicsPipeline = \
            vkCreateGraphicsPipelines(self.logicalDevice, VK_NULL_HANDLE, 1, GraphicsPipelineCreateInfo, None)[0]
        # Zrusenie Shader modulov ------------------------------------------------------------------
        vkDestroyShaderModule(self.logicalDevice, VSModule, None)
        vkDestroyShaderModule(self.logicalDevice, FSModule, None)

        # Vytvorenie CommandPool -------------------------------------------------------------------
        CommandPoolCreateInfo = VkCommandPoolCreateInfo(
            sType=VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
            queueFamilyIndex=GPUindices.graphicsFamily
        )
        self.commandPool = vkCreateCommandPool(self.logicalDevice, CommandPoolCreateInfo, None)

        # Alokovanie CommandBuffers ----------------------------------------------------------------
        CommandBufferAllocateInfo = VkCommandBufferAllocateInfo(
            sType=VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
            commandPool=self.commandPool,
            level=VK_COMMAND_BUFFER_LEVEL_PRIMARY,
            commandBufferCount=len(self.swapchainFramebuffers)
        )
        commandBuffers = vkAllocateCommandBuffers(self.logicalDevice, CommandBufferAllocateInfo)
        self.commandBuffers = [ffi.addressof(commandBuffers, i)[0] for i in range(len(self.swapchainFramebuffers))]
        for i, b in enumerate(self.commandBuffers):
            CommandBufferBeginInfo = VkCommandBufferBeginInfo(
                sType=VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
                flags=VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT
            )
            RenderPassBeginInfo = VkRenderPassBeginInfo(
                sType=VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
                renderPass=self.renderPass,
                framebuffer=self.swapchainFramebuffers[i],
                renderArea=[[0, 0], self.swapchainExtent]
            )
            clearColor = VkClearValue([[0.2, 0.2, 0.2, 1.0]])
            RenderPassBeginInfo.clearValueCount = 1
            RenderPassBeginInfo.pClearValues = ffi.addressof(clearColor)
            vkBeginCommandBuffer(b, CommandBufferBeginInfo)                           # BEGIN Command Buffer
            vkCmdBeginRenderPass(b, RenderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE)    # BEGIN Render Pass
            vkCmdBindPipeline(b, VK_PIPELINE_BIND_POINT_GRAPHICS, self.graphicsPipeline)  # BIND Graphics Pipeline
            vkCmdDraw(b, 6, 1, 0, 0)                                                      # DRAW
            vkCmdEndRenderPass(b)                                                       # END Render Pass
            vkEndCommandBuffer(b)                                                     # END Command Buffer

        # Vytvorenie Semaphore ---------------------------------------------------------------------
        SemaphoreCreateInfo = VkSemaphoreCreateInfo(
            sType=VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO
        )
        self.imageAvailableSemaphore = vkCreateSemaphore(self.logicalDevice, SemaphoreCreateInfo, None)
        self.renderFinishedSemaphore = vkCreateSemaphore(self.logicalDevice, SemaphoreCreateInfo, None)

    # Vykreslenie snimky ---------------------------------------------------------------------------
    def drawFrame(self):
        vkAcquireNextImageKHR = vkGetDeviceProcAddr(self.logicalDevice, 'vkAcquireNextImageKHR')
        vkQueuePresentKHR = vkGetDeviceProcAddr(self.logicalDevice, 'vkQueuePresentKHR')
        imageIndex = vkAcquireNextImageKHR(self.logicalDevice, self.swapchain, 18446744073709551615,
                                           self.imageAvailableSemaphore, VK_NULL_HANDLE)
        SubmitInfo = VkSubmitInfo(
            sType=VK_STRUCTURE_TYPE_SUBMIT_INFO,
            waitSemaphoreCount=1,
            pWaitSemaphores=[self.imageAvailableSemaphore, ],
            pWaitDstStageMask=[VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT, ],
            commandBufferCount=1,
            pCommandBuffers=[self.commandBuffers[imageIndex], ],
            signalSemaphoreCount=1,
            pSignalSemaphores=[self.renderFinishedSemaphore, ]
        )
        vkQueueSubmit(self.graphicsQueue, 1, SubmitInfo, VK_NULL_HANDLE)
        swapChains = [self.swapchain, ]
        PresentInfoKHR = VkPresentInfoKHR(
            sType=VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
            waitSemaphoreCount=1,
            pWaitSemaphores=[self.renderFinishedSemaphore, ],
            swapchainCount=1,
            pSwapchains=swapChains,
            pImageIndices=[imageIndex, ]
        )
        vkQueuePresentKHR(self.presentQueue, PresentInfoKHR)
        vkQueueWaitIdle(self.presentQueue)

    # Ukoncenie aplikacie --------------------------------------------------------------------------
    def exit(self):
        # Zrusenie Semaphore
        vkDestroySemaphore(self.logicalDevice, self.imageAvailableSemaphore, None)
        vkDestroySemaphore(self.logicalDevice, self.renderFinishedSemaphore, None)
        # Zrusenie CommandPool
        vkDestroyCommandPool(self.logicalDevice, self.commandPool, None)
        # Zrusenie GraphicsPipeline
        vkDestroyPipeline(self.logicalDevice, self.graphicsPipeline, None)
        # Zrusenie GraphicsPipelineLayout
        vkDestroyPipelineLayout(self.logicalDevice, self.pipelineLayout, None)
        # Zrusenie Framebuffers
        for f in self.swapchainFramebuffers:
            vkDestroyFramebuffer(self.logicalDevice, f, None)
        # Zrusenie RenderPass
        vkDestroyRenderPass(self.logicalDevice, self.renderPass, None)
        # Zrusenie ImageView
        for i in self.swapchainImageViews:
            vkDestroyImageView(self.logicalDevice, i, None)
        # Zrusenie Swapchain
        vkDestroySwapchainKHR = vkGetDeviceProcAddr(self.logicalDevice, 'vkDestroySwapchainKHR')
        vkDestroySwapchainKHR(self.logicalDevice, self.swapchain, None)
        # Zrusenie logickeho zariadenia
        vkDestroyDevice(self.logicalDevice, None)
        # Zrusenie WIN surface
        vkDestroySurfaceKHR = vkGetInstanceProcAddr(self.instance, "vkDestroySurfaceKHR")
        vkDestroySurfaceKHR(self.instance, self.surface, None)
        # Zrusenie instancie
        vkDestroyInstance(self.instance, None)
        # Zrusenie GLFW okna
        glfw.destroy_window(self.window)
        # END
        glfw.terminate()

    # Hlavny programovy cyklus ---------------------------------------------------------------------
    def mainLoop(self):
        while not glfw.window_should_close(self.window):
            glfw.poll_events()  # zachytenie udalosti
            self.drawFrame()    # vykreslenie snimky
        vkDeviceWaitIdle(self.logicalDevice)


# Instancia triedy VulkanApp
vApp = VulkanApp()

# Vytvorenie GLFW okna
vApp.createWindow()
# Inicializacia API Vulkan
vApp.initVulkan()
# Hlavny programovy cyklus
vApp.mainLoop()
# Ukoncenie aplikacie
vApp.exit()
